/* foldDb - Search database info for making foldout. */
#include "common.h"
#include "hash.h"
#include "linefile.h"
#include "obscure.h"
#include "jksql.h"
#include "psl.h"
#include "genePred.h"
#include "est3.h"
#include "gcPercent.h"
#include "rmskOut.h"
#include "agpFrag.h"
#include "cpgIsland.h"
#include "wabAli.h"
#include "clonePos.h"
#include "ctgPos.h"
#include "glDbRep.h"
#include "knownInfo.h"
#include "hdb.h"
#include "memgfx.h"

/* Coverage info file. */
char *covFile = "/projects/cc/hg/gs.3/oo.15/bed/coverage/acces_coverage";

/* Centromere/big gap info. */
char *gapFile = "centro.tab";

/* Some colors. */
struct rgbColor rgbRed = {255,0,0};
struct rgbColor rgbRedOrange = {240, 100, 0};
struct rgbColor rgbGreen = {0, 255, 0};

void usage()
/* Explain usage and exit. */
{
errAbort(
  "foldDb - Search database info for making foldout\n"
  "usage:\n"
  "   foldDb chrM chrN ...\n");
}

struct hugeGap 
/* A huge gap to be inserted. */
    {
    struct hugeGap *next;	/* Next in list. */
    int offset;                 /* Offset in ungapped chromosome. */
    int size;                   /* Size of gap. */
    char *type;                 /* Not allocated here. */
    };

int cmpHugeGap(const void *va, const void *vb)
/* Compare to sort biggest offset first. */
{
const struct hugeGap *a = *((struct hugeGap **)va);
const struct hugeGap *b = *((struct hugeGap **)vb);
return a->offset - b->offset;
}

struct chromGaps
/* A list of huge gaps on a chromosome. */
    {
    struct chromGaps *next;
    char *chrom;              /* Name of chromosome, not allocated here. */
    struct hugeGap *gapList;  /* Sorted biggest offset first. */
    };

int gapOffset(struct chromGaps *gaps, int pos)
/* Convert from ungapped to gapped position. */
{
struct hugeGap *gap;
int offset = 0;

if (gaps != NULL)
    {
    for (gap = gaps->gapList; gap != NULL; gap = gap->next)
	{
	if (pos >= gap->offset)
	    {
	    offset += gap->size;
	    pos += gap->size;
	    }
	}
    }
return offset;
}

struct hash *hugeHash;

void setupHugeGaps(struct sqlConnection *conn)
/* Setup things to lookup gaps. */
{
struct lineFile *lf;
char *words[8];
int wordCount;
struct chromGaps *chromList = NULL, *cg;
struct hugeGap *gap;
char *chrom;
char query[512];
struct sqlResult *sr;
char **row;
struct ctgPos ctgPos;
int start, size;
struct hashEl *hel;

hugeHash = newHash(6);
lf = lineFileOpen(gapFile, TRUE);
while ((wordCount = lineFileChop(lf, words)) != 0)
     {
     chrom = words[0];
     if ((cg = hashFindVal(hugeHash, chrom)) == NULL)
         {
	 AllocVar(cg);
	 slAddHead(&chromList, cg);
	 hel = hashAdd(hugeHash, chrom, cg);
	 cg->chrom = hel->name;
	 }
     size = atoi(words[3]);
     sprintf(query, "select * from ctgPos where contig = '%s'", words[2]);
     sr = sqlGetResult(conn, query);
     if ((row = sqlNextRow(sr)) == NULL)
        errAbort("Couldn't find %s from %s in database", words[2], lf->fileName);
     ctgPosStaticLoad(row, &ctgPos);
     if (!sameString(chrom, ctgPos.chrom))
         errAbort("%s is in %s in database and %s in %s", ctgPos.contig, ctgPos.chrom,
	 	chrom, lf->fileName);
     start = ctgPos.chromStart;
     uglyf("%s %s (%d size %d) %s \n", chrom, words[1], start, size, words[2]);
     sqlFreeResult(&sr);

     AllocVar(gap);
     slAddHead(&cg->gapList, gap);
     gap->offset = start;
     gap->size = size;
     gap->type = "centromere";
     if (size > 3000000)
         {
	 /* It's more than just a centromere. */
	 gap->size = 3000000;
	 size -= gap->size;
	 start += gap->size;
	 AllocVar(gap);
	 slAddHead(&cg->gapList, gap);
	 gap->offset = start;
	 gap->size = size;
	 gap->type = "other";
	 }
     }
lineFileClose(&lf);
for (cg = chromList; cg != NULL; cg = cg->next)
    slSort(&cg->gapList, cmpHugeGap);
}

void printTab(FILE *f, struct chromGaps *cg, char *chrom, int start, int end, 
	char *track, char *type, int r, int g, int b, char *val)
/* Print line to tab-delimited file. */
{
static double colScale = 1.0/255.0;
double red = r*colScale;
double green = g*colScale;
double blue = b*colScale;

if (!sameString(track, "heterochromatin"))
    {
    int offset = gapOffset(cg, (start+end)/2);
    start += offset;
    end += offset;
    }
fprintf(f, "%s\t%d\t%d\t%s\t%s\t%f,%f,%f\t%s\n",
    chrom, start, end, track, type, red, green, blue, val);
}

void printTabNum(FILE *f, struct chromGaps *cg, char *chrom, int start, int end, 
	char *track, char *type, int r, int g, int b, double num)
/* Print number-valued line to tab-delimited file. */
{
char buf[32];
sprintf(buf, "%f", num);
printTab(f, cg, chrom, start, end, track, type, r, g, b, buf);
}

void getCentrosome(struct chromGaps *cg, FILE *f)
/* Print out centrosome info (and other big gaps). */
{
struct hugeGap *hg;

for (hg = cg->gapList; hg != NULL; hg = hg->next)
    {
    struct rgbColor *col;
    if (sameString("centromere", hg->type))
         col = &rgbRed;
    else
         col = &rgbRedOrange;
    printTab(f, cg, cg->chrom, hg->offset, hg->offset+hg->size,
        "heterochromatin", "box", col->r, col->g, col->b, hg->type);
    }
}


void getKnownGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get info on known genes. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct genePred *gpList = NULL, *gp;
char geneName[64];

printf("  finding known genes\n");
sprintf(query, "select * from genieKnown where chrom = '%s' order by cdsStart", chrom);
sr = sqlGetResult(conn, query);

/* Get list of known genes. */
while ((row = sqlNextRow(sr)) != NULL)
    {
    gp = genePredLoad(row);
    slAddHead(&gpList, gp);
    }
slReverse(&gpList);
sqlFreeResult(&sr);

/* Now do query to find out name. */
for (gp = gpList; gp != NULL; gp = gp->next)
    {
    sprintf(query, "select geneName.name from knownInfo,geneName "
                   "where knownInfo.transId = '%s' "
		   "and geneName.id = knownInfo.geneName", gp->name);
    if (sqlQuickQuery(conn, query, geneName, sizeof(geneName)))
        {
	if (!sameString(geneName, "n/a"))
	    {
	    printTab(f, cg, chrom, gp->cdsStart, gp->cdsEnd, "hugo", "word", 0, 0, 150, geneName);
	    }
	}
    }
genePredFreeList(&gpList);
}

#ifdef DECENT_FAKE
void fakeKnownGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get info on known genes. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct hash *mrnaUniq = newHash(0);
struct hash *hugoUniq = newHash(0);
struct psl *pslList = NULL, *psl;
char geneName[64];
int lastStartH = -1, lastEndH = -1;

sprintf(query, "select * from %s_mrna", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    psl = pslLoad(row);
    if (psl->match + psl->repMatch >= psl->qSize/2 && !hashLookup(mrnaUniq, psl->qName))
        {
	hashAdd(mrnaUniq, psl->qName, NULL);
	slAddHead(&pslList, psl);
	}
    else
	pslFree(&psl);
    }
sqlFreeResult(&sr);

slReverse(&pslList);
for (psl = pslList; psl != NULL; psl = psl->next)
    {
    sprintf(query, "select geneName.name from mrna,geneName where mrna.geneName = geneName.id and mrna.acc = '%s'", psl->qName);
    if (sqlQuickQuery(conn, query, geneName, sizeof(geneName)))
        {
	if (!hashLookup(hugoUniq, geneName))
	    {
	    hashAdd(hugoUniq, geneName, NULL);
	    if (!sameString(geneName, "n/a") && strlen(geneName) < 8)
		{
		if (rangeIntersection(lastStartH, lastEndH, psl->tStart, psl->tEnd) <= 0)
		    {
		    printTab(f, cg, chrom, psl->tStart, psl->tEnd, 
			    "hugo", "word", 0, 0, 150, geneName);
		    }
		lastStartH = psl->tStart;
		lastEndH = psl->tEnd;
		}
	    }
	}
    }
freeHash(&mrnaUniq);
freeHash(&hugoUniq);
}
#endif /* DECENT_FAKE */

struct tickPos
/* A tick position */
    {
    struct tickPos *next;
    int start, end;
    char *val;
    };

void getPredGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f, char *table,
  int red, int green, int blue)
/* Get predicted genes. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct genePred *gp;

sprintf(query, "select * from %s where chrom = '%s' order by cdsStart", table, chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    gp = genePredLoad(row);
    printTab(f, cg, chrom, gp->cdsStart, gp->cdsEnd, 
	    "genePred", "tick", red, green, blue, ".");
    genePredFree(&gp);
    }
sqlFreeResult(&sr);
}

void getCpgIslands(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get cpgIslands. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct cpgIsland el;

sprintf(query, "select * from cpgIsland where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    cpgIslandStaticLoad(row, &el);
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "cpgIsland", "tick", 0, 120, 0, ".");
    }
sqlFreeResult(&sr);
}

void getTetHits(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get tetroadon hits. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct wabAli el;

sprintf(query, "select * from waba_tet where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    wabAliStaticLoad(row, &el);
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "tetraodon", "tick", 0, 70, 120, ".");
    }
sqlFreeResult(&sr);
}


void getEstTicks(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get est 3' ends. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct est3 *el;

sprintf(query, "select * from %s where chrom = '%s'", "est3", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    el = est3Load(row);
    printTab(f, cg, chrom, el->chromStart, el->chromEnd, 
	    "est3", "tick", 100, 50, 0, ".");
    est3Free(&el);
    }
sqlFreeResult(&sr);
}

void getGc(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Write wiggle track for GC. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct gcPercent *el;

printf("  getting GC wiggle\n");
sprintf(query, "select * from %s where chrom = '%s'", "gcPercent", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    el = gcPercentLoad(row);
    if (el->gcPpt != 0)
        {
	double scale;
	scale = (el->gcPpt - 300)*0.00300;
	if (scale > 1) scale = 1;
	if (scale < 0) scale = 0;
	printTabNum(f, cg, chrom, el->chromStart, el->chromEnd, 
		"gcPercent", "wiggle", 0, 0, 0, scale);
	}
    gcPercentFree(&el);
    }
sqlFreeResult(&sr);
}

void addSegTotals(int chromStart, int chromEnd, int *countArray, int windowSize)
/* Store the total number of bases in each window between chromStart and chromEnd
 * in the appropriate countArray element. */
{
int startWinIx = chromStart/windowSize;
int endWinIx = (chromEnd-1)/windowSize;
int winBoundary;
int winIx;
int size;

if (startWinIx == endWinIx)
    {
    countArray[startWinIx] += (chromEnd - chromStart);
    return;
    }
winBoundary = (startWinIx+1)*windowSize;
size = winBoundary - chromStart;
assert(size > 0);
assert(size <= windowSize);
countArray[startWinIx] += size;
for (winIx = startWinIx+1; winIx < endWinIx; ++winIx)
    {
    countArray[winIx] += windowSize;
    }
winBoundary = (endWinIx)*windowSize;
size = chromEnd - winBoundary;
assert(size > 0);
assert(size <= windowSize);
countArray[endWinIx] += size;
}

void getRepeatDensity(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f,
	char *repClass, int windowSize, double scale, int red, int green, int blue)
/* Put out repeat density info. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct rmskOut rmsk;
struct agpFrag frag;
int winsPerChrom = (chromSize + windowSize-1)/windowSize;
int *winBases;
int *winRepBases;
int i;
int baseCount, repCount;
double density;
double minDen = 1.0, maxDen = 0.0;
int winStart, winEnd;

printf("  Getting %s repeats in %d windows of size %d\n", repClass, winsPerChrom, windowSize);
fflush(stdout);

/* Count up repeats in each window. */
AllocArray(winRepBases, winsPerChrom);
sprintf(query, "select * from %s_rmsk where repClass = '%s'", chrom, repClass);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    rmskOutStaticLoad(row, &rmsk);
    addSegTotals(rmsk.genoStart, rmsk.genoEnd, winRepBases, windowSize);
    }
sqlFreeResult(&sr);

/* Count up bases in each window. */
AllocArray(winBases, winsPerChrom);
sprintf(query, "select * from %s_gold", chrom);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    agpFragStaticLoad(row, &frag);
    addSegTotals(frag.chromStart, frag.chromEnd, winBases, windowSize);
    }
sqlFreeResult(&sr);

/* Output density. */
for (i=0; i<winsPerChrom; ++i)
    {
    if ((baseCount = winBases[i]) > windowSize/4)
        {
	repCount = winRepBases[i];
	density = (double)repCount/(double)baseCount;
	if (density < minDen) minDen = density;
	if (density > maxDen) maxDen = density;
	density *= scale;
	if (density > 1.0) density = 1.0;
	winStart = i*windowSize;
	winEnd = winStart + windowSize;
	if (winEnd > chromSize) winEnd = chromSize;
	printTabNum(f, cg, chrom, winStart, winEnd, 
		repClass, "wiggle", red, green, blue, density);
	}
    }
printf("%s minDen %f, maxDen %f\n", chrom, minDen, maxDen);
freeMem(winBases);
freeMem(winRepBases);
}

#define finCov 100

struct rgbColor draftCols[10];


void initDraftCols()
/* Initialize draft colors with a range. */
{
int i;
int dif;
int maxCol = ArraySize(draftCols)-1;
double scale;

for (i=0; i<= maxCol; ++i)
    {
    scale = (double)i/(double)maxCol;
    dif = rgbRedOrange.r - rgbGreen.r;
    draftCols[i].r = rgbGreen.r + round(dif*scale);
    dif = rgbRedOrange.g - rgbGreen.g;
    draftCols[i].g = rgbGreen.g + round(dif*scale);
    dif = rgbRedOrange.b - rgbGreen.b;
    draftCols[i].b = rgbGreen.b + round(dif*scale);
    }
}

void makeBackupCoverageHash(struct sqlConnection *conn, struct hash *hash)
/* Make hash table that describe coverage of each clone just based on
 * whether it's draft/predraft/etc. Only do this if clone not already
 * in hash. */
{
char query[512];
struct clonePos cp;
struct sqlResult *sr;
char **row;
int coverage;
char stage;

sprintf(query, "select * from clonePos");
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    clonePosStaticLoad(row, &cp);
    stage = cp.stage[0];
    if (stage == 'D')
	coverage = 4;
    else if (stage == 'F')
	coverage = finCov;
    else if (stage == 'P')
	coverage = 2;
    else
	errAbort("Unknown stage %c for clone %s", stage, cp.name);
    chopSuffix(cp.name);
    if (!hashLookup(hash, cp.name))
	hashAdd(hash, cp.name, intToPt(coverage));
    }
sqlFreeResult(&sr);
}

void makeMainCoverageHash(char *fileName, struct hash *hash)
/* Scan file for clone coverage. */
{
struct lineFile *lf = lineFileOpen(fileName, TRUE);
char *words[16];
int wordCount;
double d;
int rd;
char *acc;

while ((wordCount = lineFileChop(lf, words)) != 0)
    {
    lineFileExpectWords(lf, wordCount, 4);
    rd = round(atof(words[3]));
    hashAddUnique(hash, words[0], intToPt(rd));
    }
lineFileClose(&lf);
}

struct hash *makeCoverageHash(struct sqlConnection *conn)
/* Get coverage hash, making it if necessary. */
{
static struct hash *hash = NULL;
if (hash == NULL)
    {
    hash = newHash(16);
    makeMainCoverageHash(covFile, hash);
    makeBackupCoverageHash(conn, hash);
    }
return hash;
}


void covInWin(struct hash *covHash, struct chromGaps *cg,
	char *chrom, int start, int end, struct gl *glList, FILE *f)
/* Process coverage in a window of chromosome. */
{
int size = end-start;
UBYTE *cov = needLargeZeroedMem(size+1);  /* A byte for each base in window. */
struct gl *gl;
int s, e;
int coverage;
char cloneName[64];
void *covAsPt;
int i,c;
int startIx, runSize;
UBYTE val, lastVal;

for (gl = glList; gl != NULL; gl = gl->next)
    {
    if ((s = gl->start) < end && (e = gl->end) > start)
        {
	if (s < start) s = start;
	if (e > end) e = end;
	assert(s<e);
	s -= start;
	e -= start;
	strcpy(cloneName, gl->frag);
	chopSuffix(cloneName);
	covAsPt = hashMustFindVal(covHash, cloneName);
	coverage = ptToInt(covAsPt);
	for (i=s; i<e; ++i)
	    {
	    c = coverage + cov[i];
	    if (c > finCov) c = finCov;
	    cov[i] = c;
	    }
	}
    }

/* Add sentinal value to simplify run-length encoding loop. */
cov[size] = finCov+1;
startIx = 0;
lastVal = cov[0];
for (i=1; i<=size; ++i)
    {
    if ((val = cov[i]) != lastVal)
        {
	if (lastVal != 0)
	    {
	    struct rgbColor *col;
	    runSize = i-startIx;
	    if (lastVal == finCov)
	        col = &rgbRed;
	    else if (lastVal >= ArraySize(draftCols))
	        col = &rgbRedOrange;
	    else
	        col = &draftCols[lastVal-1];
	    s = startIx + start;
	    e = i + start;
	    printTabNum(f, cg, chrom, s, e, 
		    "coverage", "box", col->r, col->r, col->r, lastVal);
	    }
	startIx = i;
	lastVal = val;
	}
    }
freeMem(cov);
}

void getCoverage(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f)
/* Make coverage track. */
{
struct hash *covHash = makeCoverageHash(conn);
char query[512];
struct gl *glList = NULL, *gl;
struct sqlResult *sr;
char **row;
int coverage;
int start, end;
int winSize = 1024*1024*64;

printf("  getting coverage\n");
initDraftCols();
sprintf(query, "select * from %s_gl", chrom);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    gl = glLoad(row);
    slAddHead(&glList, gl);
    }
sqlFreeResult(&sr);
slReverse(&glList);

for (start = 0; start < chromSize; start = end)
    {
    end = start + winSize;
    if (end > chromSize)
        end = chromSize;
    covInWin(covHash, cg, chrom, start, end, glList, f);
    }
glFreeList(&glList);
}

void oneChrom(char *chrom, struct sqlConnection *conn, FILE *f)
/* Get info for one chromosome.  */
{
int chromSize = hChromSize(chrom);
struct chromGaps *cg = hashFindVal(hugeHash, chrom);
printf("Processing %s\n", chrom);
if (cg != NULL)
    getCentrosome(cg, f);
#ifdef SOON
#endif SOON
getCoverage(cg, chrom, chromSize, conn, f);
getGc(cg, chrom, conn, f);
getCpgIslands(cg, chrom, conn, f);
getTetHits(cg, chrom, conn, f);
getEstTicks(cg, chrom, conn, f);
getPredGenes(cg, chrom, conn, f, "genieKnown", 0, 0, 150);
getPredGenes(cg, chrom, conn, f, "genieAlt", 100, 0, 100);
getPredGenes(cg, chrom, conn, f, "ensGene", 150, 0, 0);
getKnownGenes(cg, chrom, conn, f);
getRepeatDensity(cg, chrom, chromSize, conn, f, "SINE", 100000, 1.0/0.66, 255, 0, 0);
getRepeatDensity(cg, chrom, chromSize, conn, f, "LINE", 100000, 1.0/0.66, 0, 0, 255);
}

void foldDb(int chromCount, char *chromNames[])
/* foldDb - Search database info for making foldout. */
{
int i;
struct sqlConnection *conn = sqlConnect("hg3");

setupHugeGaps(conn);
for (i=0; i<chromCount; ++i)
    {
    char fileName[512];
    FILE *f;
    sprintf(fileName, "%s.tab", chromNames[i]);
    f = mustOpen(fileName, "w");
    oneChrom(chromNames[i], conn, f);
    fclose(f);
    }
sqlDisconnect(&conn);
}

int main(int argc, char *argv[])
/* Process command line. */
{
if (argc < 2)
    usage();
foldDb(argc-1, argv+1);
return 0;
}
